use std::ops::Deref;

use crate::api;
use crate::api::JoinResponse;
use crate::api::LoginParam;
use crate::herd_core::model::Machine;
use axum::Json;
use axum::extract::FromRequestParts;
use axum::http::StatusCode;
use axum_client_ip::ClientIp;
use axum_extra::TypedHeader;
use axum_extra::headers::Authorization;
use axum_extra::headers::authorization::Bearer;
use based::auth::Sessions;
use based::auth::User;
use owl::prelude::Model;
use owl::query;
use owl::save;
use serde::Deserialize;
use serde_json::json;
use sheepd::DeviceEntry;
use sheepd::DeviceList;

use super::mqtt::is_within_80_seconds;

pub async fn devices_list(
    session: TypedHeader<Authorization<Bearer>>,
) -> (StatusCode, Json<api::Result<DeviceList>>) {
    let machines: Vec<Model<Machine>> = query!(|_| true);

    let mut ret = vec![];

    for mac in machines {
        let id = mac.read().id.to_string().replace("-", "");
        let online_state = crate::ONLINE
            .get()
            .unwrap()
            .get(&id)
            .map(|x| is_within_80_seconds(*x.deref()))
            .unwrap_or(false);

        ret.push(DeviceEntry {
            id: id,
            hostname: mac.read().hostname.clone(),
            online: online_state,
        });
    }

    (
        StatusCode::OK,
        Json(api::Result::OkVal(DeviceList { devices: ret })),
    )
}

pub async fn login_user(Json(payload): Json<LoginParam>) -> (StatusCode, Json<serde_json::Value>) {
    log::info!("Login attempt for {}", payload.username);
    let u = User::find(&payload.username).await.unwrap();
    if u.read().verify_pw(&payload.password) {
        let ses = u.read().session().await;
        (
            StatusCode::OK,
            Json(json!(api::Result::OkVal(ses.read().token.as_str()))),
        )
    } else {
        (
            StatusCode::FORBIDDEN,
            Json(json!(api::Result::Err("invalid"))),
        )
    }
}

/// Route for joining devices (`/join`)
pub async fn join_device(
    ClientIp(ip): ClientIp,
    Json(payload): Json<api::JoinParams>,
) -> (StatusCode, Json<serde_json::Value>) {
    // TODO : check if exists already

    // TODO : validate join token
    log::info!(
        "New device joined: {} [{}]",
        payload.hostname.trim(),
        payload.machine_id
    );

    let machine = Machine::from_join_param(payload);
    let new_token = machine.token.clone();

    // TODO : add device listener instantly

    save!(machine);

    let i = crate::IDENTITY.get().unwrap();

    (
        StatusCode::OK,
        Json(json!(JoinResponse {
            token: new_token,
            identity: i.public(),
            mqtt: crate::CONFIG.get().unwrap().mqtt.clone()
        })),
    )
}
